@@ -1,34 +1,10 @@
 /*
+ * nd.c - a small program for WebDAV operations.
+ *
  * Copyright (c) 2002 Yuuichi Teranishi <teranisi@gohome.org>
+ * For license terms, see the file COPYING in this directory.
  *
- * Permission is hereby granted, free of charge, to any person obtaining
- * a copy of this software and associated documentation files (the
- * "Software"), to deal in the Software without restriction, including
- * without limitation the rights to use, copy, modify, merge, publish,
- * distribute, sublicense, and/or sell copies of the Software, and to
- * permit persons to whom the Software is furnished to do so, subject to
- * the following conditions:
- * 
- * The above copyright notice and this permission notice shall be
- * included in all copies or substantial portions of the Software.
- * 
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
- * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
- * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
- * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
- * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
- * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
- * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
-
-/*
- * nd.c : a small program for WebDAV operations.
- *
- * This module is delived from testDAV.c : a small tester program for
- * WebDAV operations which was originally written by Daniel.Veillard@w3.org,
- * that used to be included in libxml.
- */
-
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
@@ -38,6 +14,8 @@
 #include <string.h>
 #endif /* HAVE_STRING_H */
 
+#include <ctype.h>
+
 #include <libxml/xmlversion.h>
 #include <libxml/xmlmemory.h>
 #include <libxml/nanohttp.h>
@@ -66,7 +44,9 @@ ndNodeInfoFree (info)
   if (info->cdate) xmlFree(info->cdate);
   if (info->size) xmlFree(info->size);
   if (info->content) xmlFree(info->content);
-  if (info->lock) ndLockInfoFree (info->lock);
+  if (info->restype) xmlFree(info->restype);
+  if (info->props) ndPropListFree(info->props);
+  if (info->lock) ndLockInfoListFree (info->lock);
   xmlFree (info);
 }
 
@@ -100,9 +80,14 @@ ndNodeInfoPrint(fp, info, format)
       fprintf (fp, "Created: %s\n", info->cdate? info->cdate: "");
       fprintf (fp, "Size: %s\n", info->size? info->size: "");
       fprintf (fp, "Content-Type: %s\n", info->content? info->content: "");
+      fprintf (fp, "Resource-Type: %s\n", info->restype? info->restype: "");
+      if (info->props)
+	{
+	  ndPropListPrint (fp, info->props, format);
+	}
       if (info->lock)
 	{
-	  ndLockInfoPrint (fp, info->lock, format);
+	  ndLockInfoListPrint (fp, info->lock, format);
 	}
       fprintf (fp, "\n");
     }
@@ -121,10 +106,17 @@ ndNodeInfoPrint(fp, info, format)
 	fprintf (fp, " (size %s)", info->size);
       if (info->content)
 	fprintf (fp, " (content-type \"%s\")", info->content);
+      if (info->restype)
+	fprintf (fp, " (resourcetype \"%s\")", info->restype);
+      if (info->props)
+	{
+	  fprintf (fp, " ");
+	  ndPropListPrint (fp, info->props, format);
+	}
       if (info->lock)
 	{
 	  fprintf (fp, " ");
-	  ndLockInfoPrint (fp, info->lock, format);
+	  ndLockInfoListPrint (fp, info->lock, format);
 	}
       fprintf (fp, ")");
     }
@@ -156,8 +148,8 @@ ndLockInfoNew ()
 {
   ndLockInfoPtr ret = xmlMalloc (sizeof (ndLockInfo));
   if (ret == NULL) return (NULL);
-  memset (ret, 0, sizeof(ndNodeInfo));
-  return (ret);  
+  memset (ret, 0, sizeof(ndLockInfo));
+  return (ret);
 }
 
 void
@@ -173,6 +165,20 @@ ndLockInfoFree (lock)
   xmlFree (lock);
 }
 
+void
+ndLockInfoListFree (info)
+     ndLockInfoPtr info;
+{
+  ndLockInfoPtr cur = info;
+
+  while (cur != NULL)
+    {
+      info = info->next;
+      ndLockInfoFree(cur);
+      cur = info;
+    }
+}
+
 void
 ndLockInfoPrint (fp, lock, format)
      FILE *fp;
@@ -202,6 +208,126 @@ ndLockInfoPrint (fp, lock, format)
     }
 }
 
+void
+ndLockInfoListPrint (fp, info, format)
+     FILE *fp;
+     ndLockInfoPtr info;
+     int format;
+{
+  ndLockInfoPtr cur = info;
+
+  if (format == ND_PRINT_AS_SEXP)
+    fprintf (fp, "(lock-list ");
+  while (cur != NULL)
+    {
+      info = info->next;
+      ndLockInfoPrint(fp, cur, format);
+      cur = info;
+      if (format == ND_PRINT_AS_SEXP && cur != NULL) fprintf (fp, " ");
+    }
+  if (format == ND_PRINT_AS_SEXP)
+    fprintf (fp, ")\n");
+}
+
+ndPropPtr
+ndPropNew ()
+{
+  ndPropPtr ret = xmlMalloc (sizeof (ndProp));
+  if (ret == NULL) return (NULL);
+  memset (ret, 0, sizeof(ndProp));
+  return (ret);
+}
+
+void
+ndPropFree (prop)
+     ndPropPtr prop;
+{
+  if (prop == NULL) return;
+  if (prop->name) xmlFree(prop->name);
+  if (prop->value) xmlFree(prop->value);
+  if (prop->ns) xmlFree(prop->ns);
+  xmlFree (prop);
+}
+
+void
+ndPropListFree (prop)
+     ndPropPtr prop;
+{
+  ndPropPtr cur = prop;
+
+  while (cur != NULL)
+    {
+      prop = prop->next;
+      ndPropFree(cur);
+      cur = prop;
+    }
+}
+
+void
+ndPropPrint (fp, prop, format)
+     FILE *fp;
+     ndPropPtr prop;
+     int format;
+{
+  if (format == ND_PRINT_AS_HEADER)
+    {
+      if (prop == NULL) return;
+      if (prop->name)
+	{
+	  if (prop->value && *prop->value == '\"')
+	    {
+	      fprintf (fp, "Property: %s=%s", prop->name, prop->value);
+	    }
+	  else
+	    {
+	      if (prop->value)
+		fprintf (fp, "Property: %s=\"%s\"", prop->name, prop->value);
+	    }
+	}
+      if (prop->ns)
+	{
+	  fprintf (fp, "; ns=\"%s\"\n", prop->ns);
+	}
+      else
+	{
+	  fprintf (fp, "\n");
+	}
+    }
+  else if (format == ND_PRINT_AS_SEXP)
+    {
+      fprintf (fp, "(property");
+      fprintf (fp, " (name \"%s\")", prop->name? prop->name : "");
+      if (prop->value && *prop->value == '\"')
+	{
+	  fprintf (fp, " (value %s)", prop->value);
+	}
+      else
+	{
+	  fprintf (fp, " (value \"%s\")", prop->value? prop->value : "");
+	}
+      if (prop->ns)
+	fprintf (fp, " (ns \"%s\")", prop->ns);
+      fprintf (fp, ")");
+    }
+}
+
+void
+ndPropListPrint (fp, prop, format)
+     FILE *fp;
+     ndPropPtr prop;
+     int format;
+{
+  ndPropPtr cur = prop;
+
+  while (cur != NULL)
+    {
+      prop = prop->next;
+      ndPropPrint(fp, cur, format);
+      cur = prop;
+      if (format == ND_PRINT_AS_SEXP && cur != NULL) fprintf (fp, " ");
+    }
+}
+
 void*
 ndHTTPMethod (URL, auth, method, input, contentType, headers, ilen)
      const char *URL;
@@ -225,7 +351,12 @@ ndHTTPMethod (URL, auth, method, input, contentType, headers, ilen)
 
   if (auth && (auth->pauth_realm))
     {
-      sprintf (line, "Basic realm=\"%s\"", auth->pauth_realm);
+      if (snprintf (line, ND_HEADER_LINE_MAX,
+		    "Basic realm=\"%s\"", auth->pauth_realm) >=
+	  ND_HEADER_LINE_MAX)
+	{
+	  return NULL;
+	}
       if (ndAuthCreateHeader (line, auth->auth_cb, &temp_buf, 1) < 0)
 	return NULL;
       if (temp_buf == NULL) return NULL;
@@ -234,14 +365,19 @@ ndHTTPMethod (URL, auth, method, input, contentType, headers, ilen)
     }
   if (auth && (auth->auth_realm))
     {
-      sprintf (line, "Basic realm=\"%s\"", auth->auth_realm);
+      if (snprintf (line, ND_HEADER_LINE_MAX,
+		    "Basic realm=\"%s\"", auth->auth_realm) >=
+	  ND_HEADER_LINE_MAX)
+	{
+	  return NULL;
+	}
       if (ndAuthCreateHeader (line, auth->auth_cb, &temp_buf, 0) < 0)
 	return NULL;
       if (temp_buf == NULL) return NULL;
       xmlBufferAdd (header_buf, xmlBufferContent (temp_buf), -1);
       xmlBufferFree (temp_buf);
     }
-  
+
   if (headers) xmlBufferAdd (header_buf, headers, -1);
 
   ctxt = xmlNanoHTTPMethod (URL, method, input, contentType,
@@ -283,7 +419,7 @@ ndHTTPMethod (URL, auth, method, input, contentType, headers, ilen)
   if (auth && auth->notify_cb) auth->notify_cb (ctxt);
   return ctxt;
 }
-  
+
 
 /* Generic DAV request function. */
 int
@@ -303,11 +439,9 @@ nd_dav_request (method, url, auth, header, content, length,
   xmlChar buffer [1024];
   xmlBufferPtr output;
   int returnCode;
-  char *auth_header;
-  xmlBufferPtr header_buf = NULL;
 
   if (url == NULL) return -1;
-    
+
   ctxt = ndHTTPMethod(url, auth,
 		      method, content, &contentType, header,
 		      content? strlen (content): 0);
@@ -316,7 +450,7 @@ nd_dav_request (method, url, auth, header, content, length,
       return -1;
     }
   returnCode = xmlNanoHTTPReturnCode(ctxt);
-  if ((returnCode >= 300) || 
+  if ((returnCode >= 300) ||
       ((contentType != NULL) && (strncmp(contentType, "text/xml", 8)))) {
     xmlNanoHTTPClose(ctxt);
     if (contentType != NULL) xmlFree(contentType);
@@ -334,19 +468,19 @@ nd_dav_request (method, url, auth, header, content, length,
   xmlNanoHTTPClose(ctxt);
   *buf_return = output;
   return returnCode;
-}     
+}
 
 /*
- * PROPFIND
+ * PROPFIND(ALLPROP)
  */
 int
-nd_propfind_query (url, auth, depth, buf_return)
+nd_propfind_all_query (url, auth, depth, buf_return)
      char *url;
      ndAuthCtxtPtr auth;
      int depth;
      xmlBufferPtr *buf_return;
 {
-  char *verboseQuery = 
+  char *verboseQuery =
 "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\
 <D:propfind xmlns:D=\"DAV:\">\
     <D:allprop/>\
@@ -370,7 +504,7 @@ nd_propfind_query (url, auth, depth, buf_return)
   return nd_dav_request ("PROPFIND", url, auth,
 			 depth_header, verboseQuery,
 			 strlen (verboseQuery), buf_return);
-  /*  
+  /*
   if (url[strlen(url) - 1] == '/')
     return nd_dav_request ("PROPFIND", url, fn, "Depth: 1\r\n", NULL, 0,
 			   buf_return);
@@ -381,6 +515,62 @@ nd_propfind_query (url, auth, depth, buf_return)
   */
 }
 
+/*
+ * PROPFIND
+ */
+int
+nd_propfind_query (url, auth, prop, ns, depth, buf_return)
+     char *url;
+     ndAuthCtxtPtr auth;
+     char *prop;
+     char *ns;
+     int depth;
+     xmlBufferPtr *buf_return;
+{
+  char depth_header [ND_HEADER_LINE_MAX];
+  char propfind_request [ND_REQUEST_MAX];
+  char *depth_str;
+
+  if (ns == NULL)
+    {
+      ns = "DAV:";
+    }
+  switch (depth)
+    {
+    case ND_DEPTH_0:
+      depth_str = "0";
+      break;
+    case ND_DEPTH_1:
+      depth_str = "1";
+      break;
+    case ND_DEPTH_INFINITE:
+      depth_str = "Infinity";
+      break;
+    }
+  sprintf (depth_header, "Depth: %s\r\n", depth_str);
+  if (snprintf (propfind_request, ND_REQUEST_MAX,
+		"<?xml version=\"1.0\" encoding=\"utf-8\" ?>\
+<D:propfind xmlns:D=\"DAV:\">\
+  <D:prop xmlns:ns=\"%s\"><ns:%s/></D:prop>\
+</D:propfind>", ns, prop) >= ND_REQUEST_MAX)
+    {
+      return -1;
+    }
+  return nd_dav_request ("PROPFIND", url, auth,
+			 depth_header, propfind_request,
+			 strlen (propfind_request), buf_return);
+}
+
+int
+nd_dav_name_equal (node, name)
+     xmlNodePtr node;
+     char *name;
+{
+  if (node->ns && node->ns->href && (!strcmp (node->ns->href, "DAV:")))
+    return !strcmp ((char *) node->name, name);
+  return 0;
+}
+
 ndLockInfoPtr
 nd_parse_activelock (cur)
      xmlNodePtr cur;
@@ -400,7 +590,7 @@ nd_parse_activelock (cur)
 	  if (cur->children != NULL)
 	    {
 	      ret->scope = xmlMemStrdup ((char *) cur->children->name);
-	    }	  
+	    }
 	}
       else if (nd_dav_name_equal (cur, "owner"))
 	{
@@ -446,37 +636,34 @@ nd_parse_activelock (cur)
 	}
       cur = cur->next;
     }
-  return(ret);  
+  return(ret);
 }
 
 ndLockInfoPtr
 nd_parse_lockdiscovery (cur)
      xmlNodePtr cur;
 {
-  ndLockInfoPtr ret = NULL;
+  ndLockInfoPtr ret, last, res;
 
+  ret = last = NULL;
   cur = cur->children;
   while (cur != NULL)
     {
       if (nd_dav_name_equal (cur, "activelock"))
 	{
-	  ret = nd_parse_activelock (cur);
+	  res = nd_parse_activelock (cur);
+	  if (res != NULL)
+	    {
+	      if (ret == NULL) ret = last = res;
+	      else last->next = res;
+	      while (last->next != NULL) last = last->next;
+	    }
 	}
       cur = cur->next;
     }
   return ret;
 }
 
-int
-nd_dav_name_equal (node, name)
-     xmlNodePtr node;
-     char *name;
-{
-  if (node->ns && node->ns->href && (!strcmp (node->ns->href, "DAV:")))
-    return !strcmp ((char *) node->name, name);
-  return 0;
-}
-
 void
 nd_parse_prop (cur, node)
      xmlNodePtr cur;
@@ -492,6 +679,14 @@ nd_parse_prop (cur, node)
 	{
 	  node->lock = nd_parse_lockdiscovery (cur);
 	}
+      else if (nd_dav_name_equal (cur, "supportedlock"))
+	{
+	  // not supported yet.
+	}
+      else if (nd_dav_name_equal (cur, "getetag"))
+	{
+	  // not supported yet.
+	}
       else if (nd_dav_name_equal (cur, "getlastmodified"))
 	{
 	  node->date = (char *) xmlNodeGetContent(cur);
@@ -508,15 +703,37 @@ nd_parse_prop (cur, node)
 	{
 	  node->size = (char *) xmlNodeGetContent(cur);
 	}
+      else if (nd_dav_name_equal (cur, "resourcetype"))
+	{
+	  if (cur->children != NULL)
+	    {
+	      node->restype = xmlMemStrdup ((char *) cur->children->name);
+	    }
+	}
+      else if (strcmp(cur->name,"text")) /* XXX is this true? */
+	{
+	  ndPropPtr prop;
+	  prop = node->props;
+	  node->props = ndPropNew();
+	  node->props->name = xmlMemStrdup((char *) cur->name);
+	  node->props->value = (char *) xmlNodeGetContent(cur);
+	  if (cur->ns && cur->ns->href && strcmp(cur->ns->href, "DAV:"))
+	    {
+	      node->props->ns = xmlMemStrdup((char *) cur->ns->href);
+	    }
+	  // ndPropPrint (stderr, node->props, ND_PRINT_AS_SEXP);
+	  node->props->next = prop;
+	}
       cur = cur->next;
     }
 }
 
-void
+int
 nd_parse_propstat (cur, node)
      xmlNodePtr cur;
      ndNodeInfoPtr node;
 {
+  int code = -1;
   cur = cur->children;
 
   /*
@@ -530,16 +747,55 @@ nd_parse_propstat (cur, node)
 	}
       else if (nd_dav_name_equal (cur, "status"))
 	{
+	  /* Property status */
+	  char *status = (char *) xmlNodeGetContent (cur);
+	  char *cur = status;
+	  if (!strncmp(status, "HTTP/", 5))
+	    {
+	      int version = 0;
+	      int ret = 0;
+
+	      cur += 5;
+	      while ((*cur >= '0') && (*cur <= '9')) {
+		version *= 10;
+		version += *cur - '0';
+		cur++;
+	      }
+	      if (*cur == '.') {
+		cur++;
+		if ((*cur >= '0') && (*cur <= '9')) {
+		  version *= 10;
+		  version += *cur - '0';
+		  cur++;
+		}
+		while ((*cur >= '0') && (*cur <= '9'))
+		  cur++;
+	      } else
+		version *= 10;
+	      if ((*cur != ' ') && (*cur != '\t')) return -1;
+	      while ((*cur == ' ') || (*cur == '\t')) cur++;
+	      if ((*cur < '0') || (*cur > '9')) return -1;
+	      while ((*cur >= '0') && (*cur <= '9')) {
+		ret *= 10;
+		ret += *cur - '0';
+		cur++;
+	      }
+	      if ((*cur != 0) && (*cur != ' ') && (*cur != '\t'))
+		code = -1;
+	      code = ret;
+	    }
+	  xmlFree (status);
 	}
       else if (nd_dav_name_equal (cur, "responsedescription"))
 	{
 	}
       cur = cur->next;
     }
+  return code;
 }
 
 ndNodeInfoPtr
- nd_parse_prop_href(cur)
+nd_parse_prop_href(cur)
      xmlNodePtr cur;
 {
   ndNodeInfoPtr ret;
@@ -575,7 +831,10 @@ nd_parse_response(cur)
 	}
       else if (nd_dav_name_equal (cur, "propstat"))
 	{
-	  nd_parse_propstat(cur, res);
+	  if (res)
+	    {
+	      res->status = nd_parse_propstat(cur, res);
+	    }
 	}
       else if (nd_dav_name_equal (cur, "status"))
 	{
@@ -583,7 +842,6 @@ nd_parse_response(cur)
 	    {
 	      char *status = (char *) xmlNodeGetContent (cur);
 	      char *cur = status;
-
 	      if (!strncmp(status, "HTTP/", 5))
 		{
 		  int version = 0;
@@ -606,15 +864,16 @@ nd_parse_response(cur)
 		      cur++;
 		  } else
 		    version *= 10;
-		  if ((*cur != ' ') && (*cur != '\t')) return;
+		  if ((*cur != ' ') && (*cur != '\t')) return NULL;
 		  while ((*cur == ' ') || (*cur == '\t')) cur++;
-		  if ((*cur < '0') || (*cur > '9')) return;
+		  if ((*cur < '0') || (*cur > '9')) return NULL;
 		  while ((*cur >= '0') && (*cur <= '9')) {
 		    ret *= 10;
 		    ret += *cur - '0';
 		    cur++;
 		  }
-		  if ((*cur != 0) && (*cur != ' ') && (*cur != '\t')) return;
+		  if ((*cur != 0) && (*cur != ' ') && (*cur != '\t'))
+		    return NULL;
 		  res->status = ret;
 		}
 	      xmlFree (status);
@@ -625,11 +884,11 @@ nd_parse_response(cur)
 	}
       cur = cur->next;
     }
-  return(ret);
+  return (ret);
 }
 
 ndNodeInfoPtr
-nd_parse_prop_answer(cur)
+nd_parse_multistatus (cur)
      xmlNodePtr cur;
 {
   ndNodeInfoPtr ret = NULL, last, res;
@@ -653,13 +912,15 @@ nd_parse_prop_answer(cur)
 	  cur = cur->next;
 	}
     }
-  return(ret);
+  return ret;
 }
 
 int
-ndPropFind (url, auth, depth, ni_return)
+ndPropFind (url, auth, prop, ns, depth, ni_return)
      char *url;
      ndAuthCtxtPtr auth;
+     char *prop;
+     char *ns;
      int depth;
      ndNodeInfoPtr *ni_return;
 {
@@ -667,20 +928,27 @@ ndPropFind (url, auth, depth, ni_return)
   xmlDocPtr doc;
   xmlBufferPtr buf = NULL;
   int code;
-  
-  code = nd_propfind_query(url, auth, depth, &buf);
+
+  if (prop == NULL)
+    {
+      code = nd_propfind_all_query (url, auth, depth, &buf);
+    }
+  else
+    {
+      code = nd_propfind_query (url, auth, prop, ns, depth, &buf);
+    }
   if (code == -1)
     return -1;
   if (buf == NULL)
     return code;
-  /* fprintf (stderr, "%s\n", xmlBufferContent(buf)); */
+  // fprintf (stderr, "%s\n", xmlBufferContent(buf));
   doc = xmlParseMemory(xmlBufferContent(buf), xmlBufferLength(buf));
   if (doc == NULL)
     {
       xmlBufferFree (buf);
       return -1;
     }
-  ret = nd_parse_prop_answer (doc->children);
+  ret = nd_parse_multistatus (doc->children);
   xmlFreeDoc(doc);
   xmlBufferFree(buf);
   *ni_return = ret;
@@ -688,6 +956,88 @@ ndPropFind (url, auth, depth, ni_return)
   return code;
 }
 
+/*
+ * PROPPATCH
+ */
+int
+ndPropPatch (url, auth, prop, value, ns, lock_token, ni_return)
+     char *url;
+     ndAuthCtxtPtr auth;
+     char *prop;
+     char *value;
+     char *ns;
+     char *lock_token;
+     ndNodeInfoPtr *ni_return;
+{
+  char proppatch_request [ND_REQUEST_MAX];
+  char hstr [ND_HEADER_LINE_MAX];
+  ndNodeInfoPtr ret;
+  xmlBufferPtr buf = NULL;
+  xmlDocPtr doc;
+  int code;
+
+  if (ns == NULL)
+    ns = "DAV:";
+  if (lock_token)
+    {
+      if (snprintf (hstr, ND_HEADER_LINE_MAX, "If: <%s> (<%s>)\r\n",
+		    url, lock_token) >= ND_HEADER_LINE_MAX)
+	{
+	  return -1;
+	}
+    }
+  else
+    {
+      hstr[0] = '\0';
+    }
+  if (value)
+    {
+      if (snprintf (proppatch_request, ND_REQUEST_MAX,
+		    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n\
+<D:propertyupdate xmlns:D=\"DAV:\">\n\
+<D:set>\n\
+<D:prop><ns:%s xmlns:ns=\"%s\">%s</ns:%s></D:prop>\n\
+</D:set>\n\
+ </D:propertyupdate>", prop, ns, value, prop) >= ND_REQUEST_MAX)
+	{
+	  return -1;
+	}
+    }
+  else
+    { /* No value at all. REMOVE. */
+      if (snprintf (proppatch_request, ND_REQUEST_MAX,
+		    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n\
+<D:propertyupdate xmlns:D=\"DAV:\">\n\
+<D:remove>\n\
+<D:prop><ns:%s xmlns:ns=\"%s\"/></D:prop>\n\
+</D:remove>\n\
+ </D:propertyupdate>", prop, ns) >= ND_REQUEST_MAX)
+	{
+	  return -1;
+	}
+    }
+  code = nd_dav_request ("PROPPATCH", url, auth, hstr, proppatch_request,
+			 strlen (proppatch_request), &buf);
+  if (code == -1)
+    return -1;
+  if (code == 207) /* Multi status. */
+    {
+      if (buf == NULL)
+	return -1;
+      doc = xmlParseMemory(xmlBufferContent(buf), xmlBufferLength(buf));
+      if (doc == NULL)
+	{
+	  xmlBufferFree (buf);
+	  return -1;
+	}
+      ret = nd_parse_multistatus (doc->children);
+      xmlFreeDoc(doc);
+      xmlBufferFree(buf);
+      *ni_return = ret;
+    }
+  return code;
+}
+
 /*
  * PUT
  */
@@ -711,13 +1061,19 @@ ndPut (url, auth, content, length, token, ni_return)
 
   if (url == NULL) return -1;
   if_header [0] = '\0';
-  if (token) sprintf (if_header, "If: <%s> (<%s>)\r\n",
-		      url, token);
+  if (token)
+    {
+      if (snprintf (if_header, ND_HEADER_LINE_MAX,
+		    "If: <%s> (<%s>)\r\n", url, token) >= ND_HEADER_LINE_MAX)
+	{
+	  return -1;
+	}
+    }
   ctxt = ndHTTPMethod (url, auth, "PUT", content, NULL,
 		       token? if_header : NULL,
 		       length);
   if (ctxt == NULL) {
-    return(-1);
+    return -1;
   }
   code = xmlNanoHTTPReturnCode(ctxt);
   if (code == 207) /* Multi status. */
@@ -735,7 +1091,7 @@ ndPut (url, auth, content, length, token, ni_return)
 	  xmlBufferFree (buf);
 	  return -1;
 	}
-      ret = nd_parse_prop_answer (doc->children);
+      ret = nd_parse_multistatus (doc->children);
       xmlFreeDoc(doc);
       xmlBufferFree (buf);
       *ni_return = ret;
@@ -806,7 +1162,7 @@ ndPostPrint (url, auth, content, length, content_type, outfp)
   ctxt = ndHTTPMethod(url, auth, "POST", content, content_type, NULL,
 		      length);
   if (ctxt == NULL) return -1;
-  
+
   returnCode = xmlNanoHTTPReturnCode(ctxt);
   if (returnCode >= 300)
     {
@@ -819,7 +1175,7 @@ ndPostPrint (url, auth, content, length, content_type, outfp)
       fwrite (buffer, len, sizeof (xmlChar), outfp);
     }
   xmlNanoHTTPClose(ctxt);
-  return returnCode;    
+  return returnCode;
 }
 
 /*
@@ -837,8 +1193,14 @@ ndDelete (url, auth, token)
 
   if (url == NULL) return -1;
   if_header [0] = '\0';
-  if (token) sprintf (if_header, "If: <%s> (<%s>)\r\n",
-		      url, token);
+  if (token)
+    {
+      if (snprintf (if_header, ND_HEADER_LINE_MAX,
+		    "If: <%s> (<%s>)\r\n", url, token) >= ND_HEADER_LINE_MAX)
+	{
+	  return -1;
+	}
+    }
   ctxt = ndHTTPMethod (url, auth, "DELETE", NULL, NULL,
 		       token? if_header : NULL,
 		       0);
@@ -855,7 +1217,6 @@ nd_parse_lock_answer (cur)
      xmlNodePtr cur;
 {
   ndLockInfoPtr ret = NULL;
-  xmlNodePtr cur2;
 
   if (nd_dav_name_equal (cur, "prop"))
     {
@@ -870,8 +1231,8 @@ nd_parse_lock_answer (cur)
 	  cur = cur->next;
 	}
     }
-  return (ret);  
-}     
+  return (ret);
+}
 
 int
 nd_lock_request (url, auth, depth, owner, scope, timeout, buf_return)
@@ -883,7 +1244,7 @@ nd_lock_request (url, auth, depth, owner, scope, timeout, buf_return)
      char *timeout;
      xmlBufferPtr *buf_return;
 {
-  char lock_request[ND_LOCK_REQUEST_MAX];
+  char lock_request[ND_REQUEST_MAX];
   char scope_str[16];
   char hstr [ND_HEADER_LINE_MAX];
   char *depth_str;
@@ -904,15 +1265,18 @@ nd_lock_request (url, auth, depth, owner, scope, timeout, buf_return)
       sprintf (scope_str, "none");
       break;
     }
-  sprintf (lock_request,
-	   "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\
+  if (snprintf (lock_request, ND_REQUEST_MAX,
+		"<?xml version=\"1.0\" encoding=\"utf-8\" ?>\
 <D:lockinfo xmlns:D='DAV:'>\
 <D:lockscope><D:%s/></D:lockscope>\
 <D:locktype><D:write/></D:locktype>\
 <D:owner>\
 <D:href>%s</D:href>\
 </D:owner>\
-</D:lockinfo>", scope_str, owner);
+</D:lockinfo>", scope_str, owner) >= ND_REQUEST_MAX)
+    {
+      return -1;
+    }
   switch (depth)
     {
     case ND_DEPTH_0:
@@ -922,14 +1286,21 @@ nd_lock_request (url, auth, depth, owner, scope, timeout, buf_return)
     case ND_DEPTH_INFINITE:
       depth_str = "Infinity";
       break;
-    }  
+    }
   if (timeout)
-    sprintf (hstr, "Timeout: %s\r\nDepth: %s\r\n", timeout, depth_str);
+    {
+      if (snprintf (hstr, ND_HEADER_LINE_MAX,
+		    "Timeout: %s\r\nDepth: %s\r\n", timeout, depth_str)
+	  >= ND_HEADER_LINE_MAX)
+	{
+	  return -1;
+	}
+    }
   else
     sprintf (hstr, "Depth: %s\r\n", depth_str);
   return nd_dav_request ("LOCK", url, auth, hstr, lock_request,
 			 strlen (lock_request), buf_return);
-}     
+}
 
 int
 ndLock (url, auth, depth, owner, scope, timeout, li_return)
@@ -962,7 +1333,7 @@ ndLock (url, auth, depth, owner, scope, timeout, li_return)
       *li_return = ret;
     }
   return code;
-  
+
 }
 
 int
@@ -988,15 +1359,21 @@ ndUnlock (url, auth, depth, token)
     case ND_DEPTH_INFINITE:
       depth_str = "Infinity";
       break;
-    }  
+    }
 
   if (url == NULL || token == NULL) return -1;
-  sprintf (hstr, "Lock-token: <%s>\r\nDepth: %s\r\n", token, depth_str);
+  if (snprintf (hstr, ND_HEADER_LINE_MAX,
+		"Lock-token: <%s>\r\nDepth: %s\r\n", token, depth_str)
+      >= ND_HEADER_LINE_MAX)
+    {
+      return -1;
+    }
   ctxt = ndHTTPMethod (url, auth, "UNLOCK", NULL, NULL,
 		       hstr, 0);
-  if (ctxt == NULL) {
-    return(-1);
-  }
+  if (ctxt == NULL)
+    {
+      return -1;
+    }
   code = xmlNanoHTTPReturnCode(ctxt);
   xmlNanoHTTPClose (ctxt);
   return code;
@@ -1014,16 +1391,23 @@ ndMkCol (url, auth, token)
 
   if (url == NULL) return -1;
   if_header [0] = '\0';
-  if (token) sprintf (if_header, "If: <%s> (<%s>)\r\n",
-		      url, token);
+  if (token)
+    {
+      if (snprintf (if_header, ND_HEADER_LINE_MAX,
+		    "If: <%s> (<%s>)\r\n", url, token) >= ND_HEADER_LINE_MAX)
+	{
+	  return -1;
+	}
+    }
   ctxt = ndHTTPMethod (url, auth, "MKCOL", NULL, NULL,
 		       token? if_header : NULL, 0);
-  if (ctxt == NULL) {
-    return(-1);
-  }
+  if (ctxt == NULL)
+    {
+      return -1;
+    }
   code = xmlNanoHTTPReturnCode(ctxt);
   xmlNanoHTTPClose (ctxt);
-  return code;  
+  return code;
 }
 
 int
@@ -1041,12 +1425,12 @@ ndGet (url, auth, ct_return, buf_return)
   int returnCode;
 
   if (url == NULL) return -1;
-    
+
   ctxt = ndHTTPMethod(url, auth, "GET", NULL, &contentType, NULL, 0);
   if (ctxt == NULL) {
     return -1;
   }
-  
+
   returnCode = xmlNanoHTTPReturnCode(ctxt);
   if (returnCode >= 300)
     {
@@ -1064,7 +1448,8 @@ ndGet (url, auth, ct_return, buf_return)
   xmlNanoHTTPClose(ctxt);
   *buf_return = output;
   return returnCode;
-}     
+}
+
 
 int
 ndGetPrint (url, auth, ct_return, outfp)
@@ -1083,7 +1468,7 @@ ndGetPrint (url, auth, ct_return, outfp)
 
   ctxt = ndHTTPMethod(url, auth, "GET", NULL, &contentType, NULL, 0);
   if (ctxt == NULL) return -1;
-  
+
   returnCode = xmlNanoHTTPReturnCode(ctxt);
   if (returnCode >= 300)
     {
@@ -1098,7 +1483,7 @@ ndGetPrint (url, auth, ct_return, outfp)
       fwrite (buffer, len, sizeof (xmlChar), outfp);
     }
   xmlNanoHTTPClose(ctxt);
-  return returnCode;  
+  return returnCode;
 }
 
 int
@@ -1107,7 +1492,7 @@ ndMove (url, auth, dest_url, overwrite, lock_token)
      ndAuthCtxtPtr auth;
      char *dest_url;
      int overwrite;
-     char *lock_token;     
+     char *lock_token;
 {
   void *ctxt;
   char hstr [ND_HEADER_LINE_MAX*3];
@@ -1115,11 +1500,24 @@ ndMove (url, auth, dest_url, overwrite, lock_token)
   int code;
 
   if (url == NULL || dest_url == NULL) return -1;
-  sprintf (hstr, "Destination: %s\r\nOverwrite: %c\r\n", dest_url,
-	   overwrite?  'T' : 'F');
+  if (snprintf (hstr, ND_HEADER_LINE_MAX*3,
+		"Destination: %s\r\nOverwrite: %c\r\n", dest_url,
+		overwrite?  'T' : 'F') >= ND_HEADER_LINE_MAX*3)
+    {
+      return -1;
+    }
   if (lock_token)
     {
-      sprintf (hstr_1, "If: <%s> (<%s>)\r\n", dest_url, lock_token);
+      if (snprintf (hstr_1, ND_HEADER_LINE_MAX,
+		    "If: <%s> (<%s>)\r\n", dest_url, lock_token)
+	  >= ND_HEADER_LINE_MAX)
+	{
+	  return -1;
+	}
+      if ((strlen (hstr) + strlen (hstr_1) + 1) > ND_HEADER_LINE_MAX * 3)
+	{
+	  return -1;
+	}
       strcat (hstr, hstr_1);
     }
   ctxt = ndHTTPMethod (url, auth, "MOVE", NULL, NULL, hstr, 0);
@@ -1145,11 +1543,24 @@ ndCopy (url, auth, dest_url, overwrite, lock_token)
   int code;
 
   if (url == NULL || dest_url == NULL) return -1;
-  sprintf (hstr, "Destination: %s\r\nOverwrite: %c\r\n", dest_url,
-	   overwrite?  'T' : 'F');
+  if (snprintf (hstr, ND_HEADER_LINE_MAX*3,
+		"Destination: %s\r\nOverwrite: %c\r\n", dest_url,
+		overwrite?  'T' : 'F') >= ND_HEADER_LINE_MAX*3)
+    {
+      return -1;
+    }
   if (lock_token)
     {
-      sprintf (hstr_1, "If: <%s> (<%s>)\r\n", dest_url, lock_token);
+      if (snprintf (hstr_1, ND_HEADER_LINE_MAX,
+		    "If: <%s> (<%s>)\r\n", dest_url, lock_token)
+	  >= ND_HEADER_LINE_MAX)
+	{
+	  return -1;
+	}
+      if ((strlen (hstr) + strlen (hstr_1) + 1) > ND_HEADER_LINE_MAX * 3)
+	{
+	  return -1;
+	}
       strcat (hstr, hstr_1);
     }
   ctxt = ndHTTPMethod (url, auth, "COPY", NULL, NULL, hstr, 0);
@@ -1158,15 +1569,13 @@ ndCopy (url, auth, dest_url, overwrite, lock_token)
   }
   code = xmlNanoHTTPReturnCode(ctxt);
   xmlNanoHTTPClose (ctxt);
-  return code;  
+  return code;
 }
 
 char *
 ndReasonPhrase (code)
      int code;
 {
-  if (code == -1)
-    return ND_CODE_OTHER;
   switch (code)
     {
     case 100:
@@ -1262,4 +1671,5 @@ ndReasonPhrase (code)
     case 507:
       return ND_CODE_507;
     }
+  return ND_CODE_OTHER;
 }
